
from .layer import Layer

"""Map represents multiple Layers of data, intended to be combined."""
class Map:

	def __init__(self, items):
		"""Create a map from multiple named layers."""
		self.collections = dict()

		for name, item in items.items():
			self.collections[name] = Layer(item)



import pyproj

import geopandas as gpd
from tqdm import tqdm
import fiona
from fiona.crs import from_epsg
from shapely.ops import transform
from shapely.geometry import shape, mapping, MultiPolygon, box
import sys
import numpy as np

import pyproj.exceptions
import fiona.errors
import shapely.errors

if len(sys.argv) < 3:
	print('needs python3 simplify.py <input> <output>')

with fiona.open(sys.argv[1], 'r') as source:

	new_schema = source.schema
	new_schema['properties'] = {
		'EPSG': 'int',
		'name': 'str',
		'axis_unit': 'str'
	}
	new_schema['geometry'] = 'Polygon'

	with fiona.open(sys.argv[2], 'w', driver='GeoJSON',
		crs=from_epsg(4326), schema=new_schema) as out:

		for region in tqdm(source):
			if region['geometry'] is None:
				continue

			# try:
			old_prop = region['properties']
			properties = {
				'EPSG': int(old_prop['COORD_REF_SYS_CODE'] if 'COORD_REF_SYS_CODE' in old_prop else old_prop['COORD_OP_CODE']),
				'name': old_prop['COORD_REF_SYS_NAME'] if 'COORD_REF_SYS_NAME' in old_prop else old_prop['COORD_OP_NAME'],
				'axis_unit': ''
			}

			# filter out deprecated = 1
			if region['properties']['DEPRECATED'] == 1:
				continue

			# filter out for projections that exist?
			try:
				crs = pyproj.CRS('EPSG:%d' % properties['EPSG'])
				# the axis better have the same units
				if crs.axis_info[0].unit_name != crs.axis_info[1].unit_name:
					continue

				if crs.axis_info[0].unit_name in ['US survey foot', 'foot']:
					properties['axis_unit'] = 'feet'
				elif crs.axis_info[0].unit_name == 'metre':
					properties['axis_unit'] = 'meters'
				else:
					continue

			except pyproj.exceptions.CRSError:
				continue
			except Exception as e:
				raise e

			# simplify by 500m
			simplified_boundary = shape(region['geometry']).simplify(0.005, preserve_topology=False).buffer(0)

			# take largest if multipolygon
			if isinstance(simplified_boundary, MultiPolygon):
				simplified_boundary = max(simplified_boundary, key=lambda a: a.area)

			# some bounds are literally the entire globe
			# 8100 represents one quadrant of the earth
			if simplified_boundary.area > 8100:
				continue

			# round floats, way too precise by default
			rounded_boundary = shapely.wkt.loads(shapely.wkt.dumps(simplified_boundary, rounding_precision=3))

			out.write({
				'geometry': mapping(rounded_boundary),
				'properties': properties
				})



"""
mundipy.pcs automatically chooses a projected coordinate reference system
that wholly contains a given bounding box.

This is especially useful for performing operations (like distance or area)
that geographic coordinate systems are unreliable for.

The most common way this library is used is when performing operations
on unknown objects in <Latitude, Longitude> format (WGS84) with the need
to do geometric operations like distance, area, or buffering.
"""

import os

from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
import fiona

CRS_DATASET = os.path.join(os.path.dirname(__file__), 'crs_simple.fgb')

def choose_pcs(box, units='meters'):
	"""Choose a projected coordinate system from a shapely geometry with specified units for the axes."""

	if not isinstance(box, BaseGeometry):
		raise TypeError('box passed to projectedcrs.choose_pcs() is not a shapely BaseGeometry, instead got \'%s\'' % type(box))
	if units not in ['feet', 'meters']:
		raise TypeError('units passed to projectedcrs.choose_pcs() is neither feet nor meters')

	suggestions = suggest_pcs(box, units=units)
	if len(suggestions) == 0:
		raise ValueError('choose_pcs(box) had zero suggested CRS (hint: is your box in WGS84 format?)')

	return suggestions[0]

def suggest_pcs(box, units='meters', n=3):
	"""Suggest multiple projected coordinate systems from a shapely geometry with specified units for the axes."""

	if not isinstance(box, BaseGeometry):
		raise TypeError('box passed to projectedcrs.choose_pcs() is not a shapely BaseGeometry, instead got \'%s\'' % type(box))
	if units not in ['feet', 'meters']:
		raise TypeError('units passed to projectedcrs.suggest_pcs() is neither feet nor meters')

	with fiona.open(CRS_DATASET, 'r') as epsg:
		intersects_box = [geo for i, geo in epsg.items(bbox=box.bounds) if geo['geometry'] is not None]

		# filter for proprety
		has_unit = filter(lambda geo: geo['properties']['axis_unit'] == units, intersects_box)

		# filter for where items contain box
		contains_box = filter(lambda geo: shape(geo['geometry']).buffer(0).contains(box), has_unit)

		# sort by area
		satisfactory = list(map(transform_epsg, sorted(contains_box, key=lambda geo: shape(geo['geometry']).area)))

		return satisfactory[:n]

def transform_epsg(geojson):
	return {
		'name': geojson['properties']['name'],
		'epsg': geojson['properties']['EPSG'],
		'crs': 'EPSG:%d' % geojson['properties']['EPSG'],
		'units': geojson['properties']['axis_unit']
	}



from shapely.geometry.base import BaseGeometry
import shapely.wkt
import geopandas as gpd
import pandas as pd
from sqlalchemy import create_engine
from functools import lru_cache
import s2sphere

"""A Layer represents a group of spatial data."""
class Layer:

	def __init__(self, data):
		self.filename = None
		self._dataframe = None
		self._db_url = None
		self._db_table = None

		self._conn = None

		""" Initialize a Layer from a data source. """
		if isinstance(data, gpd.GeoDataFrame):
			self._dataframe = data
		elif isinstance(data, dict):
			self._db_url = data['url']
			self._db_table = data['table']
		elif isinstance(data, str):
			self.filename = data
		else:
			raise TypeError('data for Layer() is not filename, GeoDataFrame or dict with PostgreSQL details')

	def _load(self, bbox=None):
		"""Load part or the entire Layer as a dataframe."""
		if self._dataframe is not None:
			return self._dataframe

		if self._db_url is not None:
			# load from PostGIS
			self._conn = create_engine(self._db_url)

			# no bbox
			if bbox is None:
				# build the query
				query = "SELECT * FROM %s" % self._db_table

				# assume it's WGS84
				gdf = gpd.GeoDataFrame.from_postgis(query, self._conn, geom_col='geometry', crs='EPSG:4326')
				self._conn = None
				return gdf

			# tile the area
			r = s2sphere.RegionCoverer()
			# cell level 20 = typical edge length of 7-10m or 30ft
			r.max_level = 20

			p1 = s2sphere.LatLng.from_degrees(bbox[3], bbox[0])
			p2 = s2sphere.LatLng.from_degrees(bbox[1], bbox[2])

			cell_ids = r.get_covering(s2sphere.LatLngRect.from_point_pair(p1, p2))

			tiles = [self._load_tile(cellid) for cellid in cell_ids]
			together = pd.concat(tiles)
			together_geo = gpd.GeoDataFrame(data=together, geometry='geometry', crs='EPSG:4326')

			self._conn = None

			return together_geo

		if bbox is None:
			return gpd.read_file(self.filename)
		else:
			return gpd.read_file(self.filename, bbox=bbox)

	@lru_cache(maxsize=512)
	def _load_tile(self, cellid):
		vertices = [s2sphere.LatLng.from_point(s2sphere.Cell(cellid).get_vertex(v)) for v in range(4)]

		wkt = "POLYGON ((%f %f, %f %f, %f %f, %f %f, %f %f))" % (
			vertices[0].lng().degrees, vertices[0].lat().degrees,
			vertices[1].lng().degrees, vertices[1].lat().degrees,
			vertices[2].lng().degrees, vertices[2].lat().degrees,
			vertices[3].lng().degrees, vertices[3].lat().degrees,
			vertices[0].lng().degrees, vertices[0].lat().degrees,
		)
		# shapely.wkt.loads(wkt)

		# assume it's WGS84
		query = "SELECT * FROM %s WHERE geometry && ST_GeomFromEWKT('SRID=4326;%s')" % (self._db_table, wkt)

		return gpd.GeoDataFrame.from_postgis(query, self._conn, geom_col='geometry', crs='EPSG:4326')

	@property
	def dataframe(self):
		"""Load an entire Layer as a dataframe."""
		return self._load()

	"""Read into a Layer at a specific geometry (WGS84)."""
	def inside_bbox(self, bbox):
		if not isinstance(bbox, tuple) or len(bbox) != 4:
			raise TypeError('inside_bbox expected bbox to be a 4-tuple')

		return self._load(bbox=bbox)



def s2gen(max_lat, min_lat, max_lng, min_lng, s2_lvl):
    point_nw = LatLng.from_degrees(max_lat, min_lng)
    point_se = LatLng.from_degrees(min_lat, max_lng)

    rc = RegionCoverer()
    rc.min_level = s2_lvl
    rc.max_level = s2_lvl
    rc.max_cells = 1000000

    cellids = rc.get_covering(LatLngRect.from_point_pair(point_nw, point_se))

    kml = (
        '<?xml version="1.0" encoding="UTF-8"?>\n'
        '<kml xmlns="http://www.opengis.net/kml/2.2" xmlns:kml="http://www.opengis.net/kml/2.2" xmlns:atom="http://www.w3.org/2005/Atom">\n'
        '<Document><name>{0}</name>\n'
        '<Style id="s2_poly_style"><LineStyle><color>ff0000ff</color><width>2</width></LineStyle><PolyStyle><fill>0</fill></PolyStyle></Style>\n'
        '<Folder><name>{0}</name><open>1</open>'
    ).format('S2 cells level ' + str(s2_lvl))
    for cid in cellids:
        vertices = [LatLng.from_point(Cell(cid).get_vertex(v)) for v in range(4)]
        kml_coords = ['{},0'.format(swap_latlong(str(v).split()[-1])) for v in vertices]
        kml += (
            '<Placemark><name>{}</name><styleUrl>#s2_poly_style</styleUrl><Polygon><tessellate>1</tessellate><outerBoundaryIs><LinearRing>'
            '<coordinates>{} {}</coordinates>'
            '</LinearRing></outerBoundaryIs></Polygon></Placemark>'
        ).format(cid.id(), ' '.join(kml_coords), kml_coords[0])
    kml += (
        '</Folder></Document>\n'
        '</kml>'
    )

    return kml

"""VisibleLayer represents Layer data, as seen from a geometry."""
class VisibleLayer:
	# Center needs to be projected to the local CRS already
	def __init__(self, layer, bbox, center, pcs=None):
		# do local calculations in a projected CRS, which is 2D and has
		# typical distance units
		if not isinstance(pcs, str):
			raise TypeError('pcs passed to VisibleLayer() was not a string')
		self.pcs = pcs

		if not isinstance(layer, Layer):
			raise TypeError('mapdata passed to VisibleLayer() was not a mundipy.Map')
		self.local_collection = layer.inside_bbox(bbox).to_crs(crs=pcs)

		if not isinstance(center, BaseGeometry):
			raise TypeError('center passed to VisibleLayer was not shapely.geometry.BaseGeometry')
		self.center = center

	def intersects(self):
		"""Give a list of members that intersect with the center geometry."""
		spatial_index = self.local_collection.sindex
		possible_matches_index = list(spatial_index.intersection(self.center.bounds))
		possible_matches = self.local_collection.iloc[possible_matches_index]

		return possible_matches[possible_matches.intersects(self.center)]

	def nearest(self):
		"""Give the closest member to the center geometry."""
		spatial_index = self.local_collection.sindex

		idx = spatial_index.nearest(self.center.centroid)
		return self.local_collection.iloc[idx[1][0]]

	def buffer(self, distance):
		"""Buffer the geometry by some distance."""

		return self.local_collection.geometry.buffer(distance)



import json
import urllib.parse
from functools import lru_cache

from tqdm import tqdm
import geopandas as gpd
import pandas as pd
from shapely.ops import transform
from shapely.geometry import Polygon, LineString, Point, box
from shapely.geometry.base import BaseGeometry
import pyproj
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.patches as mpatches

from mundipy.map import Map
from mundipy.layer import Layer, VisibleLayer
from mundipy.api.osm import grab_from_osm
from mundipy.pcs import choose_pcs

@lru_cache(maxsize=64)
def pyproj_transform(from_crs, to_crs):
    """ Returns a pyproj transform() function between two CRS, in 'EPSG:xxxx' format."""
    return pyproj.Transformer.from_crs(pyproj.CRS(from_crs),
        pyproj.CRS(to_crs), always_xy=True).transform

class MundiQ:
    def __init__(self, center, mapdata, plot_target=None, units='meters', clip_distance=500):
        self.pcs = choose_pcs(box(*center.geometry.bounds), units=units)['crs']

        # a row in a GeoDataFrame, with a column called .geometry
        # in local projected coordinate system
        self.center = center

        to_local = pyproj_transform('EPSG:4326', self.pcs)
        self.center.geometry = transform(to_local, self.center.geometry)

        # GeoPool
        self.mapdata = mapdata

        # could be matplotlib axes or 'geojson'
        self.plot_target = plot_target
        self.plot_legend = dict()
        self.plot_handles = []
        self.plot_contents = []

        if not isinstance(clip_distance, float) and not isinstance(clip_distance, int):
            raise TypeError('clip_distance to mundi.plot() must be float or int')
        self.clip_distance = clip_distance

    """ Returns a VisibleLayer. """
    def __call__(self, *args):
        # variable length arguments
        dataset = args[0] if len(args) > 0 else None

        # First, if it's just Q(), give the center.
        if dataset == None:
            return self.center

        # Next, if dataset is in the pool, return the
        # VisibleLayer
        if dataset in self.mapdata.collections.keys():
            return VisibleLayer(self.mapdata.collections[dataset], self.bbox(), self.center.geometry, pcs=self.pcs)

        if dataset == 'openstreetmap' or dataset == 'osm':
            # Take either a tuple, or a list of tuples
            if len(args) <= 1:
                raise TypeError('openstreetmap Q() takes at least two arguments')

            # build bounding box in WGS84
            p1, p2, p3, p4 = self.bbox()

            gdf = grab_from_osm(tups=args[1], bbox=('%f,%f,%f,%f' % (p2, p1, p4, p3)))
            # convert to local CRS, because grab_from_osm gives WGS84
            return VisibleLayer(Layer(gdf), self.bbox(), self.center.geometry, pcs=self.pcs)

    def _bbox(self, distance=500):
        """Builds a bounding box around the center object in the local coordinate system."""

        return self.center.geometry.buffer(distance)

    def bbox(self, distance=500):
        """Builds a bounding box around the center object in WGS84."""
        to_wgs = pyproj_transform(self.pcs, 'EPSG:4326')
        return transform(to_wgs, self._bbox(distance=distance)).bounds

    def plot(self, shape, name):
        if self.plot_target is None:
            return

        # coerce shape into a GeoSeries, no matter what it is
        # shape can literally be anything
        if isinstance(shape, Polygon):
            shape = gpd.GeoSeries([shape])
        elif isinstance(shape, gpd.GeoDataFrame):
            shape = shape.geometry
        elif isinstance(shape, pd.Series) and isinstance(shape.geometry, BaseGeometry):
            shape = gpd.GeoSeries([shape.geometry])
        elif isinstance(shape, pd.Series) and isinstance(shape.geometry, gpd.GeoSeries):
            shape = shape.geometry
        elif isinstance(shape, VisibleLayer):
            shape = shape.local_collection
        else:
            raise TypeError('unexpected type passed to plot(), got %s' % type(shape))

        # fix self-intersections and such
        shape = shape.buffer(0)

        # clip shape by bounding box
        # especially nearby roads make the plot unreadable
        #
        shape = gpd.clip(shape, mask=self._bbox(distance=self.clip_distance))

        # LineString and Point won't show up when plotted; buffer
        shape = shape.apply(lambda g: g.buffer(2) if isinstance(g, LineString) or isinstance(g, Point) else g)

        # convert to WGS84
        shape = shape.set_crs(crs=self.pcs).to_crs(epsg=4326)

        # check color
        if name in self.plot_legend.keys():
            color = self.plot_legend[name]
        else:
            # grab next color in list
            color = list(mcolors.CSS4_COLORS.keys())[len(self.plot_legend)]
            self.plot_legend[name] = color
            self.plot_handles.append(mpatches.Patch(color=color, label=name))

        if self.plot_target == 'geojson':
            # plot_contents can only contain shapely geometries
            for idx, row in shape.items():
                self.plot_contents.append((row, mcolors.CSS4_COLORS[color]))
        else:
            shape.plot(ax=self.plot_target, edgecolor='black', facecolor=color, aspect='equal', alpha=0.75)

class Mundi:
    def __init__(self, mapdata, main: str, units='meters'):
        self.mapdata = mapdata

        self.main = self.mapdata.collections[main]

        if units not in ['meters', 'feet']:
            raise TypeError('units passed to Mundi() was neither meters nor feet')
        self.units = units

    def plot(self, fn, clip_distance=500, output_type='matplotlib'):
        if output_type == 'geojson':
            pass
        elif output_type == 'matplotlib':
            fig, ax = plt.subplots()

        # TODO: drop duplicates, except it's very slow
        #.drop_duplicates(subset=['geometry'])
        Q = MundiQ(self.main.dataframe.iloc[1], self.mapdata, plot_target=('geojson' if output_type == 'geojson' else ax), units=self.units, clip_distance=clip_distance)
        res = fn(Q)

        if output_type == 'geojson':
            # convert to dfs
            dfs = gpd.GeoDataFrame(data=Q.plot_contents, crs='EPSG:4326', columns=['geometry', 'fill'], geometry='geometry')

            # 'http://geojson.io/#data=data:application/json,%s' % urllib.parse.quote(
            return dfs.to_json()
        elif output_type == 'matplotlib':
            ax.legend(handles=Q.plot_handles, loc='upper right')
            plt.show()

    def q(self, fn, progressbar=False, n=0):
        # make iterator unique by geometry
        unique_iterator = self.main.dataframe.drop_duplicates(subset=['geometry'])

        res_keys = None
        res_outs = dict()
        # progressbar optional
        finiter = list(unique_iterator.iterrows())
        if progressbar:
            finiter = tqdm(finiter, total=len(unique_iterator))

        for idx, window in finiter:
            # fn(Q) can edit window
            original_shape = window.geometry

            Q = MundiQ(window, self.mapdata)
            res = fn(Q)

            # if res is None, skip
            if res is None:
                continue

            # coerce to tuple
            if not isinstance(res, dict):
                raise TypeError('function passed to mundi.q() must return dict or None but instead got %s' % type(res).__name__)

            if res_keys is None:
                res_keys = res.keys()

                for key in res.keys():
                    res_outs[key] = []
                res_outs['geometry'] = []
            elif res_keys != res.keys():
                raise TypeError('function passed to mundi.q() returned dict with different keys')

            for key, val in res.items():
                res_outs[key].append(val)
            res_outs['geometry'].append(original_shape)

            # check n
            if n > 0 and len(res_outs['geometry']) == n:
                break

        return gpd.GeoDataFrame(res_outs, crs='EPSG:4326', geometry='geometry')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mundipy', 'mundipy.api', 'mundipy.pcs']

package_data = \
{'': ['*']}

install_requires = \
['Fiona>=1.8.21,<2.0.0',
 'SQLAlchemy>=1.4.40,<2.0.0',
 'Shapely>=1.8.4,<2.0.0',
 'geopandas>=0.11.1,<0.12.0',
 'matplotlib>=3.6.0,<4.0.0',
 'numpy>=1.23.3,<2.0.0',
 'overpy>=0.6,<0.7',
 'pandas>=1.5.0,<2.0.0',
 'pygeos>=0.13,<0.14',
 'pyproj>=3.4.0,<4.0.0',
 's2sphere>=0.2.5,<0.3.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'mundipy',
    'version': '0.2.2',
    'description': 'mundipy is a Python framework for spatial data analysis',
    'long_description': "# [![mundi.py](docs/logo/light.svg)](https://docs.mundi.ai)\n\n[![PyPI version](https://badge.fury.io/py/mundipy.svg)](https://pypi.org/project/mundipy/) ![GitHub issues](https://img.shields.io/github/issues/BuntingLabs/mundipy) ![PyPI - License](https://img.shields.io/pypi/l/mundipy)\n\nmundipy is a Python framework for spatial data manipulation. Built on top of\n[geopandas](https://geopandas.org/en/stable/), [GDAL](https://gdal.org/),\nand [shapely](https://shapely.readthedocs.io/en/stable/manual.html), mundi.py\nprovides a useful abstraction to eliminate the hassles of spatial data.\n\n## Projected Coordinate Systems\n\nAutomatically suggests a projected coordinate system to use, given a shapely\ngeometry in WGS84.\n\nThis prioritizes coordinate systems that:\n1. totally contain the given geometry\n2. have minimal area (probably less distortion)\n3. are not deprecated\n\n```py\n>>> from mundipy.pcs import choose_pcs\n>>> from shapely.geometry import Point\n\n>>> choose_pcs(Point(-118.24, 34.052), units='feet')\n{\n    'name': 'NAD27 / California zone VII',\n    'epsg': 26799,\n    'crs': 'EPSG:26799',\n    'units': 'feet'\n}\n```\n\n## Project Roadmap\n\n- No projections needed: automatically chooses and selects a relevant CRS when doing operations\n- Automatic spatial indexing\n- Jupyter notebook native (\\_repr\\_html\\_) that doesn't explode with massive data\n- Nearest neighbor/distance queries\n- Spatial joins\n- Dissolving into h3/s2\n\n## License\n\nMundi.py is MIT licensed.\n",
    'author': 'Brendan Ashworth',
    'author_email': 'brendan@buntinglabs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://buntinglabs.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import overpy
from shapely.ops import polygonize
from shapely.geometry import shape, LineString, Point, Polygon, MultiPolygon
import geopandas as gpd

def grab_from_osm(tups=[], bbox=None):
	if not isinstance(bbox, str):
		raise TypeError('bbox is not string')

	api = overpy.Overpass()

	# tups is ['highway=residential']
	# convert to [('node', ..), ('way', ..), ('relation', ..)]
	tups = map(lambda tag: [(osmt, tag) for osmt in ['node', 'way', 'relation']], tups)
	tups = [item for sublist in tups for item in sublist]

	overpass_center = '\n'.join(list(map(lambda t: "%s[%s](%s);" % (t[0], t[1], bbox), tups)))

	result = api.query("[out:json];(%s); out body; >; out skel qt;" % overpass_center)

	return result_to_gdf(result)

def node_to_shape(n):
	return Point(n.lon, n.lat)

def way_to_shape(w):
	if len(w.get_nodes()) > 2:
		return Polygon([Point(node.lon, node.lat) for node in w.get_nodes()])
	elif len(w.get_nodes()) == 2:
		return LineString([Point(node.lon, node.lat) for node in w.get_nodes()])
	else:
		raise TypeError('way_to_shape() called on way with 1 or 0 nodes')

def result_to_gdf(results):
	nodes = results.get_elements(overpy.Node)
	node_series = map(lambda n: (node_to_shape(n),), nodes)

	ways = results.get_elements(overpy.Way)
	way_series = map(lambda w: (way_to_shape(w),), ways)

	return gpd.GeoDataFrame(data=list(node_series) + list(way_series),
		geometry='geometry', crs='EPSG:4326', columns=['geometry'])
